{

VGA256 Version 3.2 by Paul Edmondson:

NOTE: This unit was written for Turbo Pascal version 7.0. Possibly some of
      the code will NOT WORK on lower versions -- I have not tried it yet.


This unit was created by me, Paul Edmondson, in an attempt to make VGA mode
   13h a little easier to access through an initialization routine and other
   miscellaneous routines. Thanks to the ease of direct access to video
   memory, high-speed flicker-free programming is within anyone's grasp. In
   addition, I have included a 8-bit PCX viewer, along with the header format
   for anyone's use. You are free to use this unit as long as you give me
   credit for what I have done.

Speed-dependent code that is not too involved was written in assembler to
   increase performance when possible. With the help of an elementary book on
   assembly language, you should be able to understand (possibly even
   optimize) the code that I have written. Perhaps the code is repetitively
   redundant (He he) every now and then, but sometimes that's what's needed
   to best optimize the speed, or to make your code simpler to understand.
   Certain procedures do just call another procedure; if you really need to
   squeeze the speed out, you can call that procedure yourself -- but this
   only makes a noticeable difference if you call that procedure
   repetitevely. I've never claimed to be a programmer genius, and I won't
   start now.


For those of you who are not comfortible with graphics programming, here's a
   little tutorial:

Before I give you the tutorial of everything that is included in this unit, I
   need to give you the foundations this unit is based on. First, VGA memory
   starts at memory location A000:0000. Mode 13h has a resolution of 320
   pixels wide by 200 pixels high -- since each pixel takes up 8 bits or
   1 byte, the size of the mode 13h screen is 320*200*1, or 64000. Well,
   that's just under the stack size limit for Turbo Pascal. If we could just
   place an array there, we could access the video memory directly, and get
   super-fast output. I made this array, and called it videobuffer (actually
   it's a pointer to an array, so make sure it has a '^' at the end).
   Unfortunately, Turbo Pascal seems to do everything backwards, so you have
   to put the y-coordinate first in the coordinate set to turn on a pixel.
   For instance, to set (43,152) to color number 123, you would say:

       videobuffer^[152,43]:=123;

It's that easy! To clear the screen completely to color 0 (which normally is
   black), you just use the command, 'cls'.

But that's not all. Another thing about mode 13h is setting the colors to
   what you want. You see, VGA cards use sort of a look-up table for their
   colors. When you first run mode 13h, you'll see that there is already a
   set of 256 colors, but you don't have to stick with that. Although you can
   only have 256 colors on your screen at once, you can choose which 256 you
   want out of a set of 262,144 colors! The way you do that is you build the
   color out of the three primary colors of light: red, green, and blue (you
   can always experiment to see what combinations take on what colors --
   you'll get the hang of it). This system is called the RGB system of
   color. Here's how you set any color number:

       setrgbpal(89,0,0,63);

   What that does is that it sets for color number 89 the red value to 0, the
   green value to 0, and the blue value to 63, which comes out to be a bluish
   color. For each color number (0..255), you can set the red, green, and
   blue values from 0 to 63 to form one of the many possible colors. If you
   want to find out what, say the red value of color 10 is, you would access
   palette[10].r; for the green it would be palette[10].g and so on. (Please
   note that if you change the contents of this array, the colors will NOT
   change.) When you read a PCX file, the picture's RGB values are set for
   you if you choose.

But what about that flicker-free animation? Well, we already recognized that
   we could have an array the size of the screen in memory, so why don't we
   just make a fake screen array and draw everything onto the fake screen?
   Then we can just copy the fake screen to the video buffer in one fell
   swoop, and not get any flickering, because none of the drawing is done
   on the true screen. This technique is called double buffering (I recommend
   that beginners do not work with this -- it can get complicated). All we
   have to do is first make a variable of the type screenptr, and initialize
   them. So something like this:

       new(doublebuffer);

   initializes the double buffer (doublebuffer is already declared in this
   unit) and

       dispose(doublebuffer);

   deinitializes the double buffer (put at the end of the program). To set a
   pixel in the buffer is just like with the videobuffer:

       doublebuffer^[14,75]:=43;

   sets (75,14) to color 43. To access a location on the double buffer (like
   on the video buffer) you would use the expression 'doublebuffer^[14,75]'.
   To copy the double buffer to the video buffer, you would say:

       buffer(doublebuffer);

   or
       scrncopy(doublebuffer,videobuffer);

   which do the same thing. You can have more than one double buffer, but you
   will have to declare them yourself, and watch out for memory overloads.
   The scrncopy command will copy any doublebuffer to another doublebuffer as
   well, which comes in handy, for example, when a background needs to be
   replaced after every frame. The latter procedure is written in assembly to
   assure maximum performance. To flush a double buffer to empty, you would
   use the command:

       dcls(doublebuffer);

   The only problem I have found with double buffering is that you can't use
   the Turbo Pascal commands like circle or line in anything but the video
   buffer -- if you want those drawn in the double buffer, you'll have to
   make them yourself (although there is a box command). But other than that,
   the double buffering technique works quite well. (By the way, the color
   defined as empty is set by the 'colorzero' variable.


Other notes:

286 instructions

If this file has been given in its original format, then included with it
   should be a file called EGAVGACL.INC. It includes the default palettes for
   both EGA (16) and VGA (256) mode color graphics, for whatever you need
   them for. It also includes the default palette assignment for the lookup
   tables in text and 16 color graphics, for use with setpalette, getpal,
   and setallpalette.

Any of the routines included in the Turbo Pascal units Graph or Graph3 is
   completely compatible with this unit, although I would strongly recommend
   only using its procedures when you need something not included in this
   unit, due to the fact that any procedures in this unit work in any buffer
   you want and are about a billion times faster. (Well, almost.)

The palette functions, rgb, vgapalset, vgapareq, setpalette, and getpal
   all work in any VGA screen mode, not just mode 13h. That should help any
   mode 12h (640x480x16) or mode 11h (640x480x2) users. On VGA cards, palette
   commands can even be used in text mode! The palette numbers for each of
   the sixteen EGA colors is in EGAVGACL.INC, providing setpalette,
   getpal, or setallpalette have not been used.

The PCX viewer only works on certain PCX files, and is not comletely error-
   checked -- so make sure you know what you're doing with it. You can't
   really mess up anything with it -- but I have had to reboot before because
   of a bad PCX file. Just watch out.

Another note -- a lot of these procedures assume that you are already in
   graphics mode. Unlike the Turbo Pascal graphics commands which give an
   error when used in the wrong mode, the results of these are unpredictable.
   Just watch out.

I hope that this unit is error-free, as I have used it on many different
   programs of mine. If you do find an error, let me know, so that I can fix
   it. And if you find a faster way to do something, let me know that too.

The side comments on the pcxheader type are mainly for 256 (8-bit) color PCXs
   to be read by my reader.

If you have any questions about what I have done or what you want to do,
   let me know at Ityllux@aol.com.


Here's a list of all the procedures you can use in your programming:

NOTE: When I talk about 'the array palette,' I mean the array called,
      'palette' that comes with the unit. Superficially, in this case, means
      just by what's on the screen, not with the array palette.


Initializition routines:

procedure setmode(vmode:word);
   This procedure initizlizes any screen mode via a BIOS call. No TP-unit
      graphics are able, but this does not require VGA256.BGI. vmode is the
      number of the screen mode (for help on these, check out the MODE13h and
      TEXTMODE constants).

function  detectvga256:integer;
   This function is used in the initvga256 procedure -- don't worry about it.

procedure initvga256(endprog:boolean);
   This procedure initializes VGA256, or mode 13h. endprog specifies whether
      the program should end if a VGA card is not detected. If the file
      VGA256.BGI (the Turbo Pascal graphics driver for this mode) is not
      found in the current path, a path will be asked for. NOTE: This
      initilizes VGA256.BGI-dependant mode, but TP-unit graphics are enabled.

procedure closevga256(endprog:boolean);
   This closes up vga256 mode 13h and goes back to text mode. endprog is
      whether or not the program ends after the graphics are closed.


Base graphics commands:

procedure dpixel(dbuffer:screenptr; x,y:word; clr:byte);
   This sets the (x,y) in dbuffer to the given color.

procedure pixel(x,y:word; clr:byte);
   This does the same as the above, only directly to the screen.

function  gdpixel(dbuffer:screenptr; x,y:word):byte;
   This function returns the value of (x,y) in dbuffer.

function  gpixel(x,y:word):byte;
   This returns the color of the pixel at (x,y).

procedure fastbox(dbuffer:screenptr; x,y,w,h:word; colr:byte);
   This procedure draws a box in dbuffer between (x,y) and (x+w,y+h) of color
      colr.

procedure dline(dbuffer:screenptr; x1,y1,x2,y2:integer; colr:byte);
   This just draws a line in dbuffer of color colr between (x1,y1) and
      (x2,y2).


Palette manipulation commands:

procedure rgb(var triple:rgbtriple; r,g,b:byte);
   This procedure sets triple to any RGB triple, specified by r,g, and b.

procedure vgapalset;
   This superficially changes color vgapalnum to the RGB triple defined by
      vgapalr.

procedure vgapalreq;
   This set vgapalr to the superficial RGB values of color vgapalnum.

procedure setrgbpal(num,r,g,b:byte);
   This procedure sets the specified palette register, num, to the specific
      r, g, and b values. Remember that only the lower six bits of those
      values are used - so the only worthwhile values are from 0-63. This
      procedure also updates the current lookup palette set. (And for all
      you VGA-users -- this is about 100 million times as fast as
      SetRGBPalette!)

procedure getrgbpal(num:byte; var r,g,b:byte);
   This gets the on-screen RGB values for the given color (num).

procedure loadpal(lp:rgbpalette);
   This procedure loads a complete superficial palette, i.e., the array
      palette is not changed, but the on-screen palette is.

procedure loadpalette(lpal:rgbpalette);
   This procedure loads a complete palette into the screen and the lookup
      palette set.

procedure getpal(lpal:rgbpalette);
   This loads into lpal the current on-screen palette.

procedure zerooutpal;
   This superficially changes the VGA on-screen palette to all-black (i.e.,
      the array palette is not changed, but the screen is).

procedure zeropal(var lpal:rgbpalette);
   This procedure zeros out all the colors in a given palette.

procedure negpal;
   This superficially shows the reverse of the current palette.

procedure negativepal(var lpal:rgbpalette);
   This reverses lpal.

procedure grayscale;
   This superfically converts the entire picture into shades of gray.

procedure graypal(var lpal:rgbpalette);
   This converts any palette into shades of gray.

procedure palmorph(spal,dpal:rgbpalette; len,totlen:word; updatepal:boolean);
   This morphs spal into dpal. See below for the rest.

procedure fade(dir:fadetype; r,g,b:byte; len,totlen:word; updatepal:boolean);
   This superficially fades the on-screen palette to one of the following
      fade types (specified by the variable, dir): fadefrom fades from the
      specified RGB value to the current palette, fadeto fades from the
      current palette to the specified RGB value, toneg and fromneg fade to
      and from a reverse palette, and togray and fromgray fade to and from a
      grayscale picture.
   len is how long it would take to fade completely (in 64ths of a second on
      my computer -- you'll have to experiment to find out the speeds on
      yours -- it's processor-dependent), and totlen is the amount of time it
      fades (for a partial fade). updatepal determines whether the array
      palette is updated with the fade.

function  findclr(lpal:rgbpalette; r,g,b:byte):byte;
   This function returns the number of the first instance of the color with
      the closest match to the specified RGB values from lpal. If two or more
      different colors are similarly close to the given RGB values, the first
      color will be given.


PCX read routines:

procedure readpcx256(filename:string; var pbuffer:screenptr; var pal:rgbpalette; loadpal:boolean);
   This reads filename in, determines if it's a legitimate PCX file of the
      type we want, and if so, puts it into pbuffer, any screen buffer you
      want (videobuffer, doublebuffer...etc.). It puts the palette for the
      picture into pal and loads it up if loadpal is true.


Memory assignment routines:

procedure memcopy(source,destination:pointer; len:word);
   This procedure copies len number of bytes from source to destination.


Screen-change commands:

procedure scrncopy(dbuffer,nbuffer:screenptr);
   This procedure copies the contents of one screen buffer to another.

procedure buffer(dobuffer:screenptr);
   This procedure copies the contents of the given double buffer into the
      video buffer.

procedure fillbuffer(dbuffer:screenptr; fill:byte);
   This fills the given buffer with pixels of color fill.

procedure pointbuffer(dobuffer:screenptr);
   This procedure is exactly like the above procedure, except it copies to
      the video buffer more slowly, one point at a time, to give a reverse-
      dissolving sort of look.

procedure dcls(dbuf:screenptr);
   This sets all the pixels in the given buffer to 0.

procedure cls;
   This sets all the pixels in the video buffer to 0.

procedure dissolvescreen(clr:byte);
   This does exactly what it says -- the video buffer all to clr.


Sprite handling routines: (blanked out for repairs)



The variables that come with the unit are:

videobuffer,doublebuffer:screenptr;
   videobuffer is originally set to the location of the video buffer.
      doublebuffer is uninitialized.

vidbuffer,dobuffer:screen2ptr;
   vidbuffer is of the condensed screen pointer type, and is originally set
      to the location of the video buffer. dobuffer is uninitialized.

grmode:integer;
   The current graphics mode that the screen is in, although this only works
      when the VGA256.BGI is used.

palette:rgbpalette;
   The current palette -- to be looked at and not changed.

vgapalr:rgbtriple;
   Used for the procedures vgapal and vgapalr.

vgapalnum:byte;
   Used for the procedures vgapal and vgapalr.

sprite1:sprite;
   A built-in sprite that you can use.

fader,fadeg,fadeb:boolean;
   Determine whether the r, g, and b values are included in a fade.

colorzero:byte;
   Determines which color number is considered empty.

transpoints=250000;
   A constant of the number of points plotted on a dissolve of any sort.

zerotriple:rgbtriple = (r:0;g:0;b:0);
   A constant of a black color.

MODE13h=$13;
   The value for use with setmode to initialize the independant mode 13h.

TEXTM=$03;
   The value for use with setmode to return to text mode from independant
      mode 13h.

I hope that this helps -- maybe I might get my font or sprite units out soon.

Until next time....
}


unit vga256;

{$G+}

interface

type
    rgbtriple=record
                    r,g,b:byte;
              end;
    torgbtriple=^rgbtriple;
    rgbpalette=array[0..255] of rgbtriple;
    torgbpalette=^rgbpalette;
    screen=array[0..199,0..319] of byte;
    screenptr=^screen;
    screen2=array[0..31999] of byte;
    screen2ptr=^screen2;
    pcxheader=record
                    descriptor:byte;            {10 for all PCXs}
                    version:byte;               {Version number (Better be 5!)}
                    encoding:byte;              {Always 1}
                    bits_per_pixel:byte;        {Better be 8!}
                    ox,oy:integer;              {The origin x and y coord.}
                    w,h:integer;                {The width and height (Better be less than 320 and 200, respectively!)}
                    hres,vres:integer;          {The horiz. and vert. resolutions}
                    ega_palette:array[0..47] of byte; {Forget it!}
                    reserved:byte;              {Forget it!}
                    num_planes:byte;            {Number of planes (Forget it!)}
                    bytes_per_line:integer;     {How many bytes each lines takes up (Forget it!)}
                    paltype:integer;            {Got me. (Forget it!)}
                    empty:array[0..57] of byte; {Padding (Forget it!)}
              end; {ox,oy,w, and h are all that matter when loading...}
    fadetype=(fadeto,fadefrom,toneg,fromneg,togray,fromgray);
    sprite=record
                 width,height:word;
                 pic:pointer;
           end;

const
     transpoints=250000;
     zerotriple:rgbtriple = (r:0;g:0;b:0);
     MODE13h=$0013;
     TEXTM=$0003;

var
   videobuffer,doublebuffer:screenptr;
   vidbuffer,dobuffer:screen2ptr;
   grmode:integer;
   palette:rgbpalette;
   vgapalr:rgbtriple;
   vgapalnum:byte;
   sprite1:sprite;
   fader,fadeg,fadeb:boolean;
   colorzero:byte;

procedure setmode(vmode:word);
function  detectvga256:integer;
procedure initvga256(endprog:boolean);
procedure closevga256(endprog:boolean);
procedure dpixel(dbuffer:screenptr; x,y:word; clr:byte);
procedure pixel(x,y:word; clr:byte);
function  gdpixel(dbuffer:screenptr; x,y:word):byte;
function  gpixel(x,y:word):byte;
procedure fastbox(dbuffer:screenptr; x,y,w,h:word; colr:byte);
procedure dline(dbuffer:screenptr; x1,y1,x2,y2:integer; colr:byte);
procedure rgb(var triple:rgbtriple; r,g,b:byte);
procedure vgapalset;
procedure vgapalreq;
procedure setrgbpal(num,r,g,b:byte);
procedure getrgbpal(num:byte; var r,g,b:byte);
procedure loadpal(lp:rgbpalette);
procedure loadpalette(lpal:rgbpalette);
procedure getpal(var lpal:rgbpalette);
procedure zerooutpal;
procedure zeropal(var lpal:rgbpalette);
procedure negpal;
procedure negativepal(var lpal:rgbpalette);
procedure grayscale;
procedure graypal(var lpal:rgbpalette);
procedure palmorph(spal,dpal:rgbpalette; len,totlen:word; updatepal:boolean);
procedure fade(dir:fadetype; r,g,b:byte; len,totlen:word; updatepal:boolean);
function  findclr(lpal:rgbpalette; r,g,b:byte):byte;
procedure readpcx256(filename:string; var pbuffer:screenptr; var pal:rgbpalette; loadpal:boolean);
procedure memcopy(source,destination:pointer; len:word);
procedure scrncopy(dbuffer,nbuffer:screenptr);
procedure buffer(dobuffer:screenptr);
procedure fillbuffer(dbuffer:screenptr; fill:byte);
procedure pointbuffer(dobuffer:screenptr);
procedure dcls(dbuf:screenptr);
procedure cls;
procedure dissolvescreen(clr:byte);

implementation

uses crt,graph;

const
     vmask=$3C6;
     vread=$3C7;
     vwrite=$3C8;
     vdata=$3C9;

var
   topal,tovgapalr,tovgapalnum:pointer;


{Initializition routines}

procedure setmode(vmode:word); assembler;

asm
   mov ax,word ptr vmode
   int 10h
end;


function detectvga256:integer;

var
  ddriver:integer;
  smode:integer;

begin
  detectgraph(ddriver, smode);
  if (ddriver=VGA) or (ddriver=MCGA) then
    detectvga256:=0
  else
    detectvga256:=grerror;
end;


procedure initvga256(endprog:boolean);

var
  autodetect:pointer;
  grdriver,errcode,lcv:integer;
  path:string;

begin
     directvideo:=true;
     path:='';
     repeat
           autodetect:=@detectvga256;
           grdriver:=installuserdriver('VGA256', autodetect);
           grdriver:=detect;
           initgraph(grdriver, grmode, path);
           errcode:=graphresult;
           if errcode<>grok then
              begin
                   writeln('Graphics Error: ', grapherrormsg(errcode));
                   if errcode=grfilenotfound then
                      begin
                           writeln('Enter full path to BGI driver or type <ctrl-break> to quit:');
                           readln(path);
                           writeln;
                      end
                   else
                       if endprog then
                          halt(1);
              end;
     until errcode=grok;
     for lcv:=0 to 255 do
         setpalette(lcv,lcv);
end;


procedure closevga256(endprog:boolean);

begin
     closegraph;
     if endprog then
        halt(1);
end;


{Base graphics commands}

procedure dpixel(dbuffer:screenptr; x,y:word; clr:byte); assembler;

asm
   cmp  x,320
   jge  @@1
   cmp  y,200
   jge  @@1
   push di
   les  di,dbuffer
   shl  y,6
   add  di,y
   shl  y,2
   add  di,y
   add  di,x
   mov  al,byte ptr clr
   mov  es:[di],al
   pop  di
@@1:
end;


procedure pixel(x,y:word; clr:byte);

begin
     dpixel(videobuffer,x,y,clr);
end;


function gdpixel(dbuffer:screenptr; x,y:word):byte;

begin
     if (x<320) and (y<200) then
        gdpixel:=dbuffer^[y,x]
     else
         gdpixel:=0;
end;


function gpixel(x,y:word):byte;

begin
     gpixel:=gdpixel(videobuffer,x,y);
end;


procedure fastbox(dbuffer:screenptr; x,y,w,h:word; colr:byte); assembler;

asm
   push bx
   cmp  x,320
   jge  @@1
   mov  bx,x
   add  bx,w
   cmp  bx,320
   jge  @@1
   cmp  y,200
   jge  @@1
   mov  bx,y
   add  bx,h
   cmp  bx,200
   jge  @@1
   push di
   les  di,dbuffer
   shl  y,6
   add  di,y
   shl  y,2
   add  di,y
   add  di,x
   mov  cx,h
@@2:
   push di
   push cx
   mov  cx,w
   mov  al,colr
   mov  ah,al
   rep  stosw
   pop  cx
   pop  di
   add  di,320
   loop @@2;
   pop  di
@@1:
   pop  bx
end;


procedure dline(dbuffer:screenptr; x1,y1,x2,y2:integer; colr:byte);

var
   lcv:word;
   slice:word;
   curx,cury:longint;

begin
     if abs(x2-x1)<abs(y2-y1) then
        slice:=abs(y2-y1)
     else
         slice:=abs(x2-x1);
     for lcv:=0 to slice do
         begin
              curx:=((x2-x1)*lcv) div slice+x1;
              cury:=((y2-y1)*lcv) div slice+y1;
              dpixel(dbuffer,curx,cury,colr);
         end;
end;


{Palette manipulation commands}

procedure rgb(var triple:rgbtriple; r,g,b:byte);

begin
     triple.r:=r;
     triple.g:=g;
     triple.b:=b;
end;


procedure vgapalset; assembler;

asm
   push di
   push ax
   mov  ax,$FF
   mov  dx,vmask
   out  dx,ax
   les  di,tovgapalnum
   mov  al,es:[di]
   mov  dx,vwrite
   out  dx,al
   les  di,tovgapalr
   mov  dx,vdata
   mov  cx,3
@@1:
   mov  al,es:[di]
   out  dx,al
   add  di,1
   loop @@1
   pop  ax
   pop  di
end;


procedure vgapalreq; assembler;

asm
   push di
   push ax
   mov  ax,$FF
   mov  dx,vmask
   out  dx,ax
   les  di,tovgapalnum
   mov  al,es:[di]
   mov  dx,vread
   out  dx,al
   les  di,tovgapalr
   mov  dx,vdata
   mov  cx,3
@@1:
   in   al,dx
   mov  es:[di],al
   add  di,1
   loop @@1
   pop  ax
   pop  di
end;


procedure setrgbpal(num,r,g,b:byte); assembler;

asm
   push di
   push ax
   mov  ax,$FF
   mov  dx,vmask
   out  dx,ax
   mov  al,num
   mov  dx,vwrite
   out  dx,al
   mov  dx,vdata
   mov  al,r
   out  dx,al
   mov  al,g
   out  dx,al
   mov  al,b
   out  dx,al
   les  di,topal
   mov  ax,word ptr num
   add  di,ax
   shl  ax,2
   add  di,ax
   mov  al,byte ptr r
   mov  es:[di],al
   add  al,byte ptr g
   mov  di,1
   mov  es:[di],al
   add  di,1
   mov  al,byte ptr b
   mov  es:[di],al
   pop  ax
   pop  di
end;


procedure getrgbpal(num:byte; var r,g,b:byte);

var
   vpr:rgbtriple;

begin
     vgapalnum:=num;
     vgapalreq;
     r:=vpr.r;
     g:=vpr.g;
     b:=vpr.b;
end;


procedure loadpal(lp:rgbpalette);

begin
     for vgapalnum:=0 to 255 do
         begin
              vgapalr:=lp[vgapalnum];
              vgapalset;
         end;
end;


procedure loadpalette(lpal:rgbpalette);

begin
     palette:=lpal;
     loadpal(lpal);
end;


procedure getpal(var lpal:rgbpalette);

begin
     for vgapalnum:=0 to 255 do
         begin
              vgapalreq;
              lpal[vgapalnum]:=vgapalr;
         end;
end;


procedure zerooutpal;

begin
     vgapalnum:=colorzero;
     vgapalreq;
     for vgapalnum:=0 to 255 do
         vgapalset;
end;


procedure zeropal(var lpal:rgbpalette);

var
   pla:rgbpalette;

procedure zerop(tolpal:torgbpalette); assembler;

asm
   push di
   push ax
   les  di,tolpal
   mov  cx,256
@@1:
   mov  al,0
   push cx
   mov  cx,3
@@2:
   mov  es:[di],al
   add  di,1
   loop @@2
   pop  cx
   loop @@1
   pop  ax
   pop  di
end;

begin
     zerop(@pla);
     lpal:=pla;
end;


procedure negpal;

begin
     for vgapalnum:=0 to 255 do
         begin
              vgapalr.r:=63-palette[vgapalnum].r;
              vgapalr.g:=63-palette[vgapalnum].g;
              vgapalr.b:=63-palette[vgapalnum].b;
              vgapalset;
         end;
end;


procedure negativepal(var lpal:rgbpalette);

var
   lcv:byte;

begin
     for lcv:=0 to 255 do
         begin
              lpal[lcv].r:=63-lpal[lcv].r;
              lpal[lcv].g:=63-lpal[lcv].g;
              lpal[lcv].b:=63-lpal[lcv].b;
         end;
end;


procedure grayscale;

begin
     for vgapalnum:=0 to 255 do
         begin
              vgapalr.r:=(palette[vgapalnum].r+palette[vgapalnum].g+
                          palette[vgapalnum].b) div 3;
              vgapalr.g:=vgapalr.r;
              vgapalr.b:=vgapalr.r;
              vgapalset;
         end;
end;


procedure graypal(var lpal:rgbpalette);

var
   lcv,temp:byte;
   pt:rgbtriple;

begin
     for lcv:=0 to 255 do
         begin
              pt:=palette[lcv];
              temp:=(pt.r+pt.g+pt.b) div 3;
              lpal[lcv].r:=temp;
              lpal[lcv].g:=temp;
              lpal[lcv].b:=temp;
         end;
end;


procedure palmorph(spal,dpal:rgbpalette; len,totlen:word; updatepal:boolean);

var
   lcv:word;
   clr:byte;

procedure updatecolors;

var
   h:byte;

begin
     for vgapalnum:=0 to 255 do
         begin
              if fader then
                 begin
                      h:=spal[vgapalnum].r;
                      vgapalr.r:=round((dpal[vgapalnum].r-h)*(lcv/len)+h);
                 end;
              if fadeg then
                 begin
                      h:=spal[vgapalnum].g;
                      vgapalr.g:=round((dpal[vgapalnum].g-h)*(lcv/len)+h);
                 end;
              if fadeb then
                 begin
                      h:=spal[vgapalnum].b;
                      vgapalr.b:=round((dpal[vgapalnum].b-h)*(lcv/len)+h);
                 end;
              vgapalset;
         end;
end;

begin
     if len=0 then
        zerooutpal
     else
         for lcv:=0 to totlen do
             updatecolors;
     if updatepal then
        getpal(palette);
end;


procedure fade(dir:fadetype; r,g,b:byte; len,totlen:word; updatepal:boolean);

var
   lcv:byte;
   pal1,pal2,temp:rgbpalette;

begin
     if (dir=fadeto) or (dir=fadefrom) then
        for lcv:=0 to 255 do
            begin
                 temp[lcv].r:=r;
                 temp[lcv].g:=g;
                 temp[lcv].b:=b;
            end;
     case dir of
          fadeto:begin
                      pal1:=palette;
                      pal2:=temp;
                 end;
          fadefrom:begin
                        pal1:=temp;
                        pal2:=palette;
                   end;
          toneg:begin
                     pal1:=palette;
                     pal2:=palette;
                     negativepal(pal2);
                end;
          fromneg:begin
                       pal1:=palette;
                       negativepal(pal1);
                       pal2:=palette;
                  end;
          togray:begin
                      pal1:=palette;
                      pal2:=palette;
                      graypal(pal2);
                 end;
          fromgray:begin
                        pal1:=palette;
                        graypal(pal1);
                        pal2:=palette;
                   end;
     end;
     palmorph(pal1,pal2,len,totlen,updatepal);
end;


function findclr(lpal:rgbpalette; r,g,b:byte):byte;

var
   closeness,temp:word;
   closest:byte;
   lcv:byte;

begin
     closeness:=256;
     for lcv:=0 to 255 do
         begin
              temp:=abs(r-lpal[lcv].r)+abs(g-lpal[lcv].g)+abs(b-lpal[lcv].b);
              if temp<closeness then
                 begin
                      closeness:=temp;
                      closest:=lcv;
                 end;
         end;
     findclr:=closest;
end;


{PCX read routines}

procedure readpcx256(filename:string; var pbuffer:screenptr; var pal:rgbpalette; loadpal:boolean);

var
   pcxhf:file of pcxheader;
   pcxf:text;
   header:pcxheader;
   lcv,curx,cury:integer;
   inb:char;
   inbyte,count:byte;

begin
     assign(pcxhf,filename);
     reset(pcxhf);
     read(pcxhf,header);
     close(pcxhf);
     if (header.descriptor=10) and (header.encoding=1) and (header.version=5) and (header.bits_per_pixel=8)
                               and (header.ox+header.w<=320) and (header.oy+header.h<=200) then
       begin
             if pbuffer=nil then
                begin
                     new(pbuffer);
                     dcls(pbuffer);
                end;
             assign(pcxf,filename);
             reset(pcxf);
             for lcv:=1 to 128 do
                 read(pcxf,inb);
             cury:=header.oy;
             repeat
                   curx:=header.ox;
                   repeat
                         read(pcxf,inb);
                         inbyte:=byte(inb);
                         if inbyte>=192 then
                            begin
                                 count:=inbyte-192;
                                 read(pcxf,inb);
                                 inbyte:=byte(inb);
                                 for lcv:=1 to count do
                                     begin
                                          dpixel(pbuffer,curx,cury,inbyte);
                                          inc(curx);
                                     end;
                            end
                         else
                             begin
                                  dpixel(pbuffer,curx,cury,inbyte);
                                  inc(curx);
                             end;
                   until curx>header.w-header.ox;
                   inc(cury);
             until cury>header.h-header.oy;
             read(pcxf,inb);
             inbyte:=byte(inb);
             if inbyte=12 then
                begin
                     for lcv:=0 to 255 do
                         begin
                              read(pcxf,inb);
                              inbyte:=byte(inb);
                              pal[lcv].r:=inbyte shr 2;
                              read(pcxf,inb);
                              inbyte:=byte(inb);
                              pal[lcv].g:=inbyte shr 2;
                              read(pcxf,inb);
                              inbyte:=byte(inb);
                              pal[lcv].b:=inbyte shr 2;
                              if loadpal then
                                 begin
                                      vgapalnum:=lcv;
                                      vgapalr:=pal[lcv];
                                      vgapalset;
                                 end;
                         end;
                end;
             close(pcxf);
        end;
end;


{Memory assignment routines}

procedure memcopy(source,destination:pointer; len:word); assembler;

asm
   push di
   push ds
   les  di,destination
   lds  si,source
   mov  cx,len
   rep  movsb
   pop  ds
   pop  di
end;


{Screen-change commands}

procedure scrncopy(dbuffer,nbuffer:screenptr); assembler;

asm
   push di
   push ds
   les  di,nbuffer
   lds  si,dbuffer
   mov  cx,32000
   rep  movsw
   pop  ds
   pop  di
end;


procedure buffer(dobuffer:screenptr);

begin
     scrncopy(dobuffer,videobuffer);
end;


procedure pointbuffer(dobuffer:screenptr);

var
   count:longint;
   x,y:integer;

begin
     for count:=1 to transpoints do
         begin
              x:=random(320);
              y:=random(200);
              pixel(x,y,dobuffer^[y,x]);
         end;
     scrncopy(dobuffer,videobuffer);
end;


procedure fillbuffer(dbuffer:screenptr; fill:byte); assembler;

asm
   push di
   les  di,dbuffer
   mov  cx,32000
   mov  al,fill
   mov  ah,al
   rep  stosw
   pop  di
end;


procedure dissolvescreen(clr:byte);

var
   lcv:longint;

begin
     for lcv:=1 to transpoints do
         videobuffer^[random(200),random(320)]:=clr;
     fillbuffer(videobuffer,clr);
end;


procedure dcls(dbuf:screenptr);

begin
     fillbuffer(dbuf,colorzero);
end;


procedure cls;

begin
     fillbuffer(videobuffer,colorzero);
end;


{Sprite handling routines}


{Done when the program loads}

begin
     videobuffer:=ptr($A000,$0000);
     vidbuffer:=ptr($A000,$0000);
     doublebuffer:=nil;
     dobuffer:=nil;
     sprite1.pic:=nil;
     topal:=@palette;
     tovgapalr:=@vgapalr;
     tovgapalnum:=@vgapalnum;
     fader:=true;
     fadeg:=true;
     fadeb:=true;
     colorzero:=0;
end.
